#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef __APPLE__
  #include <mach/error.h>
#else
  #include <error.h>
#endif
#include <errno.h>

#ifndef THRESH
#define THRESH 128
#endif

static int debug_level = 0;

static int max3(int a, int b, int c) {
  if ((a > b) && (a > c)) return a;
  if (b > c) return b;
  return c;
}

#define SP ' '
#define PX '@'
#define XS SP // '.'
#define NO SP // 'C'

typedef struct CHAR {
  char *output;
  char *fname;
  char *desc;
} CHAR;

int main(int argc, char **argv) {
  int i, j;
  const CHAR *f[128];
  for (i = 0; i < 128; i++) f[i] = NULL;

  // This is an initial phase of recognition where one glyph is manually
  // chosen by inspection from the html map, to be used for initial
  // identification.

  // We will then use that to collect multiple instances of each glyph
  // (which will almot certainly require manual filtering to remove
  //  misidentified matches).
  // Those can then be used for two purposes:
  // 1) training a neural net recogniser
  // 2) composing an 'ideal' glyph by averaging multiple (aligned)
  //    instances of a glyph.  I'm hoping that this will remove a
  //    lot of the noise and produce a cleaner master.  Manual setting
  //    of the threshold level may be needed to extract the best image.

  // one of each character manually entered from first page for demo
  // These initial approximations are replaced later by ideal characters
  // created by merging multiple instances of those characters.

  int next = 0;
  // At this point the empty cells are handled slightly differently
  // and the space character has to be listed here first.
  f[next++] = & (CHAR) { " ", "../testdata/IPA/tiles/IPA-02_1_1/line001/col001.png", "space" };
  f[next++] = & (CHAR) { "n", "../testdata/IPA/tiles/IPA-02_1_1/line008/col007.png", "n" };
  f[next++] = & (CHAR) { "h", "../testdata/IPA/tiles/IPA-02_1_1/line030/col007.png", "h" };
  f[next++] = & (CHAR) { "ˈ", "../testdata/IPA/tiles/IPA-02_1_1/line008/col002.png", "IPA Primary stress" };
  f[next++] = & (CHAR) { "æ", "../testdata/IPA/tiles/IPA-02_1_1/line008/col003.png", "IPA Ash" };
  f[next++] = & (CHAR) { "l", "../testdata/IPA/tiles/IPA-02_1_1/line008/col004.png", "l" };
  f[next++] = & (CHAR) { "p", "../testdata/IPA/tiles/IPA-02_1_1/line008/col005.png", "p" };
  f[next++] = & (CHAR) { "ə", "../testdata/IPA/tiles/IPA-02_1_1/line008/col006.png", "schwa" };
  f[next++] = & (CHAR) { "e", "../testdata/IPA/tiles/IPA-02_1_1/line038/col012.png", "e" };
  f[next++] = & (CHAR) { "ɜ", "../testdata/IPA/tiles/IPA-02_1_1/line025/col006.png", "backwards epsilon" };
  f[next++] = & (CHAR) { "s", "../testdata/IPA/tiles/IPA-02_1_1/line008/col009.png", "s" };
  f[next++] = & (CHAR) { "t", "../testdata/IPA/tiles/IPA-02_1_1/line008/col010.png", "t" };
  f[next++] = & (CHAR) { "ɒ", "../testdata/IPA/tiles/IPA-02_1_1/line008/col011.png", "upside down a (rounded a)" };
  f[next++] = & (CHAR) { "k", "../testdata/IPA/tiles/IPA-02_1_1/line008/col012.png", "k" };
  f[next++] = & (CHAR) { "f", "../testdata/IPA/tiles/IPA-02_1_1/line009/col005.png", "f" };
  f[next++] = & (CHAR) { "b", "../testdata/IPA/tiles/IPA-02_1_1/line010/col007.png", "b" };
  f[next++] = & (CHAR) { "ɛ", "../testdata/IPA/tiles/IPA-02_1_1/line010/col008.png", "IPA Epsilon" };
  f[next++] = & (CHAR) { "ɪ", "../testdata/IPA/tiles/IPA-02_1_1/line011/col011.png", "Unicode lax I" };
  f[next++] = & (CHAR) { ",", "../testdata/IPA/tiles/IPA-02_1_1/line011/col014.png", "," };
  f[next++] = & (CHAR) { "-", "../testdata/IPA/tiles/IPA-02_1_1/line011/col016.png", "-" };
  f[next++] = & (CHAR) { "a", "../testdata/IPA/tiles/IPA-02_1_1/line012/col010.png", "a" };
  f[next++] = & (CHAR) { "z", "../testdata/IPA/tiles/IPA-02_1_1/line012/col012.png", "z" };
  f[next++] = & (CHAR) { "d", "../testdata/IPA/tiles/IPA-02_1_1/line012/col016.png", "d" };
  f[next++] = & (CHAR) { "ɔ", "../testdata/IPA/tiles/IPA-02_1_1/line015/col002.png", "Open o" };
  f[next++] = & (CHAR) { "o", "../testdata/IPA/tiles/IPA-02_1_1/line016/col006.png", "o" };
  f[next++] = & (CHAR) { "ʊ", "../testdata/IPA/tiles/IPA-02_1_1/line016/col007.png", "IPA Lax u (upside down omega)" };
  f[next++] = & (CHAR) { "ə̲", "../testdata/IPA/tiles/IPA-02_1_1/line020/col006.png", "underlined schwa" };
  f[next++] = & (CHAR) { "e̲", "../testdata/IPA/tiles/IPA-02_1_1/line048/col006.png", "underlined e" };
  f[next++] = & (CHAR) { "ʃ", "../testdata/IPA/tiles/IPA-02_1_1/line021/col011.png", "IPA long Esh" };
  f[next++] = & (CHAR) { "/", "../testdata/IPA/tiles/IPA-02_1_1/line025/col011.png", "/" };
  f[next++] = & (CHAR) { "v", "../testdata/IPA/tiles/IPA-02_1_1/line029/col011.png", "v" };
  f[next++] = & (CHAR) { "ˌ", "../testdata/IPA/tiles/IPA-02_1_1/line030/col006.png", "IPA Secondary stress" };
  f[next++] = & (CHAR) { "ð", "../testdata/IPA/tiles/IPA-02_1_1/line031/col005.png", "IPA Eth" };
  f[next++] = & (CHAR) { "m", "../testdata/IPA/tiles/IPA-02_1_1/line032/col008.png", "m" };
  f[next++] = & (CHAR) { "j", "../testdata/IPA/tiles/IPA-02_1_1/line033/col008.png", "j" };
  f[next++] = & (CHAR) { "u", "../testdata/IPA/tiles/IPA-02_1_1/line033/col009.png", "u" };
  f[next++] = & (CHAR) { "g", "../testdata/IPA/tiles/IPA-02_1_1/line036/col008.png", "g" };
  f[next++] = & (CHAR) { "r", "../testdata/IPA/tiles/IPA-02_1_1/line037/col009.png", "r" };
  f[next++] = & (CHAR) { "ʌ", "../testdata/IPA/tiles/IPA-02_1_1/line044/col005.png", "central vowel" };
  f[next++] = & (CHAR) { "w", "../testdata/IPA/tiles/IPA-02_1_1/line048/col005.png", "w" };
  f[next++] = & (CHAR) { ";", "../testdata/IPA/tiles/IPA-02_1_1/line049/col005.png", ";" };
  f[next++] = & (CHAR) { "i", "../testdata/IPA/tiles/IPA-02_1_1/line056/col018.png", "i" };
  f[next++] = & (CHAR) { "θ", "../testdata/IPA/tiles/IPA-02_1_1/line057/col009.png", "theta" };
  f[next] = NULL;
  
  printf("#ifndef NULL\n");
  printf("#define NULL 0\n");
  printf("#endif\n\n");

  printf("typedef struct asciimap {\n");
  printf("  const char *ch;\n"); // updated structure required for IPA example.
                                 // Backporting to initial implementation and updating ALGOL-W example.
  printf("  const char *f;\n");
  printf("  const char *desc;\n");
  printf("  const char *line[100]; // *UP TO* 45 lines.\n");
  printf("} asciimap;\n\n");

  printf("const asciimap C[128] = {\n");
  for (i = 0; i < next; i++) {
    printf("  {\n");
    printf("    \"%s\",\n", f[i]->output);
    printf("    \"%s\",\n", f[i]->fname);
    printf("    \"%s\",\n", f[i]->desc);
    {
      FILE *png;
      static char fname[1024];
      sprintf(fname, "convert %s txt:-", f[i]->fname);
      png = popen(fname, "r");
      if (png == NULL) {
        printf("      {\n");
        //for (j = 0; j < 45; j++) {
        //  printf("    \"                                                      \",\n");
        //}
        printf("        NULL,\n");
        printf("      },\n");
      } else {
        static char line[128];
        int c;
        int maxx, maxy, maxbyte, x, y, r, g, b;
        // # ImageMagick pixel enumeration: 40,67,255,gray
        // 0,0: (255,255,255)  #FFFFFF  gray(255)
        int p;
        printf("    {\n");
        for (;;) {
          p = 0;          
          for (;;) {
            c = fgetc(png);
            if (c == EOF || c == '\n' || ferror(png)) break;
            line[p++] = c;
          }
          line[p] = '\0';
          if (c != '\n') break;
          if (line[0] == '#') {
            // "# ImageMagick pixel enumeration: 40,67,255,gray"
            int rc = sscanf(line, "# ImageMagick pixel enumeration: %d,%d,%d,"/*gray*/,
                            &maxx, &maxy, &maxbyte);
            if (rc == 3) {
              if (debug_level >= 1) fprintf(stderr, "%d x %d (%d)\n", maxx, maxy, maxbyte);
              //P.width = maxx; P.height = maxy;
            } else {
              exit(1); 
            }
          } else {
            // "0,0: (255)  #FFFFFF  gray(255)"
            int rc = sscanf(line, "%d,%d: (%d,%d,%d)"/*  #FFFFFF  gray(255)*/, &x, &y, &r, &g, &b);
            if (rc == 5) {
              if (debug_level >= 2) printf("x=%d y=%d: r=%d/g=%d/b=%d\n", x, y, r, g, b);
              if ((x >= maxx || x < 0) || (y >= maxy || y < 0)) {
                // index out of range
                if (debug_level >= 2) fprintf(stderr, "png2h: pixel (%d,%d) out of range (0:%d,0:%d):  %s\n\n", x,y, maxx-1,maxy-1, line);
              } else {
                if (x == 0) printf("      \"");
                if (max3(r,g,b) > THRESH) putchar('@'); else putchar(' ');
                if (x == maxx-1)  printf("\",\n");
              }
            } else {
              rc = sscanf(line, "%d,%d: (%d)"/*  #FFFFFF  gray(255)*/, &x, &y, &g);
              if (rc == 3) {
                if (debug_level >= 2) printf("x=%d y=%d: g=%d\n", x, y, g);
                if ((x >= maxx || x < 0) || (y >= maxy || y < 0)) {
                  // index out of range
                  if (debug_level >= 2) fprintf(stderr, "png2h: pixel (%d,%d) out of range (0:%d,0:%d):  %s\n\n", x,y, maxx-1,maxy-1, line);
                } else {
                  if (x == 0) printf("      \"");
                  if (g > THRESH) putchar('@'); else putchar(' ');
                  if (x == maxx-1)  printf("\",\n");
                }
              } else {
                exit(1);
              }
            }
          }
          
        }
        fclose(png);
        printf("      NULL,\n    },\n");
      }
    }
    printf("  },\n");
  }
  printf("};\n");
}
